/* INC-GEN.c (version 3). This is an implementation of the INC-GEN algorithm from A. Arman, P. Gao and N. Wormald,
 * Fast uniform  generation of random graphs with given degree sequences, 2019 IEEE 60th Annual Symposium on
 * Foundations of Computer Science (FOCS) 1, 1371-1379 (2019).
 * The algorithm generates a uniformly random graph with given degree sequence.
 * Original code written by Alon Naor. Enquiries: nicholas.wormald@monash.edu
 */

/*
 Version information:
 
 2020/08/09 v2: improved i/o
 2020/08/25 v3: Graphicality check bug fixed
 
 */
 
/******************************************************************************************************************/
/********************************************   LICENCE INFORMATION   *********************************************/
/******************************************************************************************************************/

/*
 
 Copyright 2020 Monash University
 
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at
 
 http://www.apache.org/licenses/LICENSE-2.0
 
 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
 
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <stdbool.h>
#include <string.h>
#include <sys/time.h>
#include <inttypes.h>
#include <limits.h>



/******************************************************************************************************************/
/**********************************************   USER DEFINITIONS   **********************************************/
/******************************************************************************************************************/

/*********************************   RNG  *********************************/
/* Define EXACTLY one of the following to select the (pseudo) Random Number Generator (RNG).
 * 1. Define LONG_CYCLE_RNG to use the algorithm from Press et al.
 * 2. Define SIMPLE_RNG to use sequential calls to C standard function rand.
 */
//#define SIMPLE_RNG
#define LONG_CYCLE_RNG

/* If USE_FIXED_SEED is set to 1, initial seed is FIXED_SEED, otherwise it is random.
 * Note that SIMPLE_RNG requires a non-negative seed, whereas
 * LONG_CYCLE_RNG is only initialized if the seed is non-positive.
 */
#define USE_FIXED_SEED 0
#define FIXED_SEED 1234
/**************************************************************************/



/***************************   GRAPHS OUTPUT   ****************************/
// Set DO_NOT_OUTPUT to 1 if you are not interested in the actual output of the generated graphs.
#define DO_NOT_OUTPUT 0

/* Determines the graph output type.
 * If OUTPUT_GRAPH_BY_VERTEX is set to 0, the output is a list of all edges in the graph.
 * Otherwise, for each vertex its list of neighbors of larger index is printed.			*/
#define OUTPUT_GRAPH_BY_VERTEX 1

/* Determines whether the graph output is sorted or not.
 * Applies for both edge list and neighbors lists output. */
#define OUTPUT_GRAPH_SORTED 1

// For neighbors lists output, determines whether vertices with empty lists are printed.
#define PRINT_EMPTY_LISTS 0

// For edge list output, determines how many edges are printed per line.
#define EDGES_PER_LINE 10

/* For edge list output, determines the column width of the output.
 * When set to 0, width is calibrated automatically according to n. */
#define COLUMN_WIDTH 0
/**************************************************************************/



/*****************************   PERFORMANCE   ****************************/
/* The hash size of any vertex is its degree * HASH_MULTIPLIER (rounded up).
 * The smaller this value is, less memory is allocated and larger graphs can be handled,
 * but more collisions will occur in the hash, resulting in slower insert/find operations.
 * The value can be any positive real number.
 */
#define HASH_MULTIPLIER 1

/* Determines the data type used for M, the sum of degrees (which affects the
 * amount of allocated memory). Set to 1 if M >= 2^31, otherwise set to 0.
 */
#define M_64 0

/* Adds safety checks such as data structures verification and brute
 * count validation for the values computed for backwards rejection.
 * Significantly slows down runtime. Should normally be set to 0.
 */
#define SAFETY_CHECKS 0

/* When DEBUG_PRINTS is set to 1, debug data is printed to log_file.
 * Should normally be set to 0, unless working with a small input and
 * wish to examine carefully the execution of the code.
 */
#define DEBUG_PRINTS 0
/**************************************************************************/



/*********************************   MISC   *******************************/
/* Sets the limit for a potential exit point after many consecutive
 * failed attempts. Set to 0 if not interested in this feature.	 */
#define CONSECUTIVE_FAILED_ATTEMPTS_LIMIT 1000

// Determines if statistics summary is printed at the end of the run.
#define PRINT_SUMMARY 0

// Determines if the counter of successfully generated graphs is printed during the run.
#define SHOW_PROGRESS 0
/**************************************************************************/

/******************************************************************************************************************/
/****************************************   OTHER DEFINITIONS (NON-USER)  *****************************************/
/******************************************************************************************************************/

typedef uint64_t u64;
#define S16_MAX  0x7fff
#define S32_MAX  0x7fffffff

/**********************************   RNG   *******************************/
/* randInt uses repeated calls to a base RNG function and concatenates random bits.
 * Here the base call and number of bits it generates are defined according to the selected RNG.
 */
#ifdef SIMPLE_RNG
#if (RAND_MAX == S16_MAX)
// Here rand() returns 15 random bits.
#define RAND_BASE_CALL   rand()
#define NUM_OF_RAND_BITS 15
#endif
#if (RAND_MAX == S32_MAX)
// Here rand() returns 31 random bits, from which the LSB is removed.
#define RAND_BASE_CALL   (rand() / 2)
#define NUM_OF_RAND_BITS 30
#endif
// Verifies the maximum value returned by rand() is supported.
#ifndef RAND_BASE_CALL
#error Cannot use SIMPLE_RNG. Use LONG_CYCLE_RNG or add support for the value of RAND_MAX.
#endif
#endif

#ifdef LONG_CYCLE_RNG
// LRAN(1 << 30) returns 30 random bits (can't have 31 since LRAN never returns INT_MAX).
#define RAND_BASE_CALL   LRAN(1 << 30)
#define NUM_OF_RAND_BITS 30
#endif

// The maximal range of integers randInt can produce is 2^60 (i.e. 60 random bits).
#define RAND_INT_MAX (1ULL << 60)

// Preprocessor verification: enforces the definition of exactly one RNG.
#if((defined(LONG_CYCLE_RNG) && defined(SIMPLE_RNG)) || \
	(!defined(LONG_CYCLE_RNG) && !defined(SIMPLE_RNG)))
#error EXACTLY one of LONG_CYCLE_RNG and SIMPLE_RNG must be defined.
#endif
/**************************************************************************/



/*****************************   PERFORMANCE   ****************************/
// Defines the type of M and the matching print specifier according to user's selection.
#if M_64
typedef long long type_M;
#define PRINT_M PRId64
#else
typedef int type_M;
#define PRINT_M PRId32
#endif /* M_64 */

#if SAFETY_CHECKS
// Makes the data structures verification calls invoke the actual functions.
#define checkPointsArray()		_checkPointsArray()
#define verifyDataStructures()  _verifyDataStructures()
#else
// Removes from the code all data structures verification calls.
#define checkPointsArray()
#define verifyDataStructures()
#endif /* SAFETY_CHECKS */
/**************************************************************************/



/***************************   OUTPUT / LOGGING   *************************/
// Determines the correct outputGraph function to be invoked.
#if DO_NOT_OUTPUT
#define outputGraph()
#elif OUTPUT_GRAPH_SORTED
#define outputGraph outputGraphSorted
#elif OUTPUT_GRAPH_BY_VERTEX
#define outputGraph outputGraphUnsortedNeighbors
#else
#define outputGraph outputGraphUnsortedEdges
#endif

// A default log file name if one is not provided.
#define DEFAULT_LOG_FILE "INC_GEN_log.txt"

// The different types of output in the code.
#define LOG_ERR(param...)	fprintf(stderr, param)
#define LOG_INFO(param...)	fprintf(log_file, param)
#define OUTPUT(param...)	fprintf(output_file, param)

#if DEBUG_PRINTS
// Enables debug logging.
#define LOG_DEBUG			 LOG_INFO
// Makes the print calls invoke the actual printing functions.
#define printPairing		 _printPairing
#define printNeighborsByHash _printNeighborsByHash
#define printAllNeighbors 	 _printAllNeighbors
#else
// Removes from the code all printing calls.
#define LOG_DEBUG(param...)
#define printPairing()
#define printNeighborsByHash(u)
#define printAllNeighbors()
#endif /* DEBUG_PRINTS */
/**************************************************************************/



/******************************************************************************************************************/
/********************************************   STRUCTURES AND ENUMS  *********************************************/
/******************************************************************************************************************/

/* <adjNode> (adjacency node)
 * Adjacency nodes are used to store the neighbors of each vertex, loops excluded.
 * For each vertex, all of its nodes are stored consecutively, and they are additionally mapped into a hash table,
 * where each hash cell contains an unordered linked list of nodes.
 * The multiplicity of the edge (single / double) corresponding to a given node is indicated by the "edge" field.
 *
 * Fields:
 * edge - this field has two functionalities, depending on the phase the algorithm is at.
 * 		  When a new pairing is generated and the related data structures are being created, points to
 * 		  the location in the points array of the corresponding edge, as long as it not detected as double.
 * 		  Once a double edge is detected, this field is used to locate the other copy of this edge, and then
 * 		  it is set to M for "double edge" indication in all four corresponding nodes.
 * 		  After all data structures have been created and during the calls to noLoops and noDoubles, this field is used
 * 		  to detect double edges nodes: "edge == M" if and only if the node corresponds to a double edge.
 * id 	- the index of the neighbor.
 * next - points to the next node in the linked list (NULL if this node is last in the list).
 */
typedef struct _adjNode{
	type_M 			 edge;
	int 			 id;
	struct _adjNode* next;
}adjNode;



/* <vertex>
 * The vertex structure holds all information of a given vertex.
 * The "has_loop", "single_edges_num" and "next_available_node" fields change according
 * to the current status of the graph, while the other fields are fixed throughout the run.
 *
 * Fields:
 * degree				- the degree of the vertex.
 * has_loop				- a boolean indicating whether there is a loop on this vertex or not.
 * single_edges_num 	- the number of single edges incident with the vertex.
 * hash_cell_size		- the number of indices mapped to each cell in the hash table. If this value is k, then the
 * 					  	  vertices in the range [ik, ..., (i + 1)k) are mapped to the i-th hash cell (including i = 0).
 * hash					- pointer to the beginning of the hash table allocated for this vertex.
 * neighbors_pool		- pointer to the beginning of the neighbors pool allocated for this vertex.
 * next_available_node	- pointer to the next available adjacency node to be allocated for this vertex.
 * 					  	  Initially points to the beginning of the section kept for this vertex in the
 * 					  	  adjacency nodes pool, then advancing with the allocation of each node.
 */
typedef struct{
	short	  degree;
	bool 	  has_loop;
	short	  single_edges_num;
	int		  hash_cell_size;
	adjNode** hash;
	adjNode*  neighbors_pool;
	adjNode*  next_available_node;
}vertex;


/* <nd_pair>
 * This structure is used to store the degree sequence given to the program as input during the initialization phase.
 *
 * Fields:
 * d	- a degree to appear in the graph.
 * n	- the number of vertices to have this degree.
 */
typedef struct{
	int d;
	int n;
}nd_pair;



// Indicates the exit reason upon program termination.
typedef enum{
	successful_termination,
	user_termination,
	file_failure,
	input_failure,
	configuration_failure,
	memory_allocation_failure,
	safety_check_failure
}exit_value;



/******************************************************************************************************************/
/**********************************************   GLOBAL VARIABLES   **********************************************/
/******************************************************************************************************************/

/* General graph parameters, depending only on the degree sequence. */
int n; 		// Number of vertices in the graph.
u64	Delta;	// Maximum degree in the graph.
			// Bounded by 2^16, but kept as u64 to avoid potential overflows in some computations.

u64 M;		// Sum of degrees.
u64 M2;		// Sum of deg(v) * (deg(v) - 1) over all v in V.
int B1;		// The maximum allowed number of loops.
int B2;		// The maximum allowed number of double edges.


/* General graph parameters which change according to the current status of the processed graph. */
int loop_count;		// Current number of loops.
int	double_count;	// Current number of double edges.
u64 P2;				// Current number of simple ordered 2-paths v1-v2-v3 such that there is no loop on v2.


/* The data structures holding all information for each processed graph. */
vertex 	*V;				// The array of all vertex structures.
int 	*points;		// The points array, stores all pairs. First loops, then single edges, then double edges.
adjNode **hash_array;	// The array of all hash entries. Distributed among the vertices according to their degrees.
adjNode	*adjNode_pool;	// The array of all adjacency nodes. Distributed among the vertices according to their degrees.

u64 total_hash_cells;	// The total number of hash cells allocated for all vertices.


/* Data structures used when computing B_d(G) to avoid double counting of "bad" configurations. */
int *C6_array;		// Used when computing B_d(G) to avoid double counting of C_6 copies.
int	*encountered;	// Used for efficient initialization of C6_array at the end of each computation of B_d(G).

/* File streams. */
FILE *output_file;	// The file to which generated graphs are printed. Set to stdout if no output file is provided.
FILE *log_file;		// The file to which the log is printed. Only applicable when DEBUG_PRINTS is set to 1.


/* RNG seed. */
int seed;			// Only applicable if LONG_CYCLE_RNG is defined. Given as a parameter in every call to LRAN.


/* Statistics variables. */
int many_loops, double_loop, many_double_edges, triple_edge;		// Counts pairing generations failures.

int total_L_switch_calls, total_D_switch_calls;						// Counts all calls to L/D-switches.
int loop_fwd_rej, loop_bck_rej, double_fwd_rej, double_bck_rej;		// Counts all f/b-rejections according to L/D.

int total_noloops,   total_loop_num;			// Counts calls to NoLoops and the accumulated sum of loops.
int total_nodoubles, total_double_edges_num;	// Counts calls to NoDoubles and the accumulated sum of double edges.



/******************************************************************************************************************/
/***************************************************   MACROS   ***************************************************/
/******************************************************************************************************************/

// Basic utility functions.
#define IS_EVEN(num) ((num) % 2 == 0)
#define FALL_2(num)  (((u64) num) * (num - 1))

// Checks if a given string is a number. Used for input parsing.
#define NOT_A_NUMBER(str) (strspn(str, "0123456789") < strlen(str))

// Compares two pairs of vertices, used when data structures are verified.
#define ARE_PAIRS_EQUAL(a1, b1, a2, b2)	(((a1 == a2) && (b1 == b2)) || ((a1 == b2) && (b1 == a2)))

/* Swaps two entries in the points array. Only invoked while generating
 * a new pairing and before P2 is computed so P2 is used as tmp.	 */
#define POINTS_SWAP(a,b){	\
	P2		  = points[a];	\
	points[a] = points[b];	\
	points[b] = P2;			\
}

// Allocates memory and exits upon failure.
#define ALLOCATE_MEMORY(name, type, num){												\
	name = NULL;																		\
	if(SIZE_MAX / sizeof(type) >= num)													\
		name = (type*) malloc (sizeof(type) * (num));									\
	if(!name){																			\
		LOG_ERR("Could not allocate memory for %s (%" PRIu64 " bytes). Exiting.\n",		\
				#name, ((u64) sizeof(type)) * (num));									\
		finalize(memory_allocation_failure);											\
	}																					\
	LOG_DEBUG("Allocated %" PRIu64 " bytes for %s.\n",									\
				((u64) sizeof(type)) * (num), #name);									\
}

// Allocates memory and sets the entire block to zero.
#define ALLOCATE_INITIALIZED_MEMORY(name, type, num){	\
	ALLOCATE_MEMORY(name, type, num);					\
	memset(name, 0, sizeof(type) * (num));				\
}

// Measures and prints the runtime of a given operation (could be a sequence of operations).
#define MEASURE(str,operation){								\
	struct timeval  tv1, tv2;								\
	gettimeofday(&tv1, NULL);								\
	operation;												\
	gettimeofday(&tv2, NULL);								\
	LOG_INFO("Total time %s = %.3f seconds.\n", str,		\
			 (tv2.tv_usec - tv1.tv_usec) / 1000000.0 +		\
			 tv2.tv_sec - tv1.tv_sec);						\
}

// Retrieves the cell in the hash of 'v' to which 'neighbor' is mapped.
#define GET_CELL(v,neighbor) (V[v].hash + ((neighbor)/V[v].hash_cell_size))

// Retrieves the adjNode of 'neighbor' from the hash of 'v' if exists, and NULL otherwise.
#define GET_NODE(v,neighbor,node){					\
	node = *(GET_CELL(v,neighbor));					\
	while(node != NULL && node->id != neighbor)		\
		node = node->next; 							\
}

// Inserts the node of a new neighbor of a vertex u to the top of its hash cell.
#define HEAD_INSERT(u, node){					\
	adjNode** head = GET_CELL(u,node->id);		\
	node->next = *head;							\
	*head = node;								\
}

// Extracts the adjacency node of 'neighbor' from the hash of 'v'.
#define EXTRACT_NODE(v, neighbor, node){		\
	/* Find the previous node in the list. */	\
	adjNode **prev = GET_CELL(v, neighbor);		\
	while((*prev)->id != neighbor){				\
		prev = &((*prev)->next);				\
	}											\
	/* Extract the node. */						\
	node  = *prev;								\
	*prev = node->next;							\
}

// Queries regarding the existence or the multiplicity of an edge are wrapped in one of the following.
#define IS_EDGE(u,v)		(countEdgeMultiplicity(u,v)  > 0)
#define IS_NO_EDGE(u,v)		(countEdgeMultiplicity(u,v) == 0)
#define IS_SINGLE_EDGE(u,v) (countEdgeMultiplicity(u,v) == 1)
#define IS_DOUBLE_EDGE(u,v) (countEdgeMultiplicity(u,v) == 2)

/* Checks the multiplicity of the edge corresponding to the given node.
 * Mainly used for neighbors iterations in which only single edges should be processed. */
#define IS_NODE_OF_SINGLE_EDGE(node) (node->edge < M)
#define IS_NODE_OF_DOUBLE_EDGE(node) (node->edge == M)

// Used when computing b_l(G) and b_d(G) for backward rejection.
#define NOT_IN_S(v) (v != S[0] && v!= S[1] && v!= S[2])

// Points to the start and the end of the allocated adjacency nodes of a given vertex.
#define START_OF_POOL(v) (V[v].neighbors_pool)
#define END_OF_POOL(v)	 (V[v].next_available_node)

// The offsets in the points array of the single and double edges sections.
#define START_OF_SINGLE_EDGES (2 * loop_count)
#define START_OF_DOUBLE_EDGES (M - 4 * double_count)

/* Chooses a random loop or a random single/double edge.
 * - RANDOM_LOOP points to the left cell of the loop edge in the points array.
 * - A random single/double edge contains the edge orientation.				*/
#define RANDOM_LOOP			(randInt(loop_count)) * 2
#define RANDOM_SINGLE_EDGE 	randInt(START_OF_DOUBLE_EDGES - START_OF_SINGLE_EDGES) + START_OF_SINGLE_EDGES
#define RANDOM_DOUBLE_EDGE 	randInt(4 * double_count) + START_OF_DOUBLE_EDGES

/* Points a randomly selected single edge to its left cell in the points array.
 * Sets the values for p1, p2 according to the original random orientation.  */
#define ADJUST_EDGE(p1,p2,edge){	\
	p1 = points[edge];				\
	if(IS_EVEN(edge)){				\
		p2 = points[edge + 1];		\
	}								\
	else{							\
		edge--;						\
		p2 = points[edge];			\
	}								\
}

// Printing formats for the statistics printout functions.
#define PERCENTAGE(a,b) 	 (b) > 0 ? 100.0 * (a) / (b) : 0
#define VAL_AND_PERCENT(a,b) a, PERCENTAGE(a, b)

/* Prints a summary of all attempts to generate a simple graph.
 * Invoked at the end of the run and after many consecutive failures. */
#define PRINT_OVERALL_STATS(mode){															\
	LOG_##mode("\nOverall:\n"																\
			   "---------\n"																\
			   "Total attempts %d.\n"														\
			   "Pairing generation failures %d (%.1f%%), NoLoops failures %d (%.1f%%), "	\
			   "NoDoubles failures %d (%.1f%%), success %d (%.1f%%).\n\n",					\
			   total_attempts, VAL_AND_PERCENT(generation_fails, total_attempts),			\
			   VAL_AND_PERCENT(noLoops_fails,total_attempts),								\
			   VAL_AND_PERCENT(noDoubles_fails,total_attempts),								\
			   VAL_AND_PERCENT(ii - 1, total_attempts));									\
}



/******************************************************************************************************************/
/**************************************************   FUNCTIONS   *************************************************/
/******************************************************************************************************************/

///////////////////////////////////////////////////////////////////
//LONGRAN generator (taken from Press et al)

#ifdef LONG_CYCLE_RNG

//#include <cmath>
#define IM1 2147483563
#define IM2 2147483399
#define AM (1.0/IM1)
#define IMM1 (IM1-1)
#define IA1 40014
#define IA2 40692
#define IQ1 53668
#define IQ2 52774
#define IR1 12211
#define IR2 3791
#define NTAB 32
#define NDIV (1+IMM1/NTAB)
#define EPS 1.2e-7
#define RNMX (1.0-EPS)
float ran2(int *idum) /* Adjusted: it's in range [0,1] */
{
	int j;
	long k;
	static long idum2=123456789;
	static long iy=0;
	static long iv[NTAB];
	float temp;

	if(*idum <= 0)
	{                            /* initialise  */
		if (-(*idum)<1) *idum=1; /* preventing idum=0 */
		else *idum = -(*idum);
		idum2 = (*idum);
		for (j=NTAB+7;j>=0;j--)
		{                        /*load the shuffle table (after 8 warm-ups) */
			k = (*idum)/IQ1;
			*idum = IA1*(*idum-k*IQ1)-k*IR1;
			if (*idum < 0) *idum += IM1;
			if (j< NTAB) iv[j] = *idum;
		}
		iy = iv[0];
	}

	k=(*idum)/IQ1;            /* start here when not initialising */
	*idum = IA1*(*idum - k*IQ1)-k*IR1; /* compute idum = (IA1*idum) %IM1 without   */
	if (*idum < 0) *idum += IM1;       /*   overflows by Schrage's method         */
	k = idum2/IQ2;
	idum2 = IA2*(idum2-k*IQ2) - k*IR2;  /* compute idum2 = (IA2*idum) %IM2 lokewise */
	if (idum2 < 0) idum2 += IM2;
	j = iy/NDIV;                       /* will be in range 0 .. NTAB - 1 */
	iy = iv[j] - idum2;                /* Here idum is shuffled, idum and idum2 are */
	iv[j] = *idum;                     /* combined to generate output */
	if(iy < 1) iy += IMM1;
	/*      if ((temp=AM*iy) > RNMX) return RNMX;   In original, this included as users don't expect endpoint value   */
	if ((temp=AM*iy) > 1) return 1.0;
	else  return temp;
}

int PRESSRAN(int n)
{
	double almost;
	almost = 1-1.0e-12;
	return floor(almost*ran2(&seed)*(n));
}

int LRAN(int n)
{
	double almost,rest,ndub;
	int rootn;
	ndub=n;
	almost = 1-1.0e-14;
	rootn=trunc( sqrt(ndub));
	rest=(ndub)/(rootn*1.0);
	return trunc(almost*(PRESSRAN(rootn)+ ran2(&seed))*rest );
}

#endif /* LONG_CYCLE_RNG */

// End of LONGRAN generator.
////////////////////////////////////////////////////////////////////

// Frees allocated memory and closes output file(s) before terminating.
void finalize(exit_value val){

	// Free allocated memory (some pointers may be NULL pointers).
	free(V);
	free(points);
	free(adjNode_pool);
	free(hash_array);
	free(encountered);
	free(C6_array);

	LOG_DEBUG("All memory freed.\n");

	// Close open files.
	fclose(log_file);
	if(output_file != stdout)
		fclose(output_file);

	exit(val);
}

/* Returns a random integer between 0 and (range - 1). Implementation depends on the RNG selected.
 * In case range is small enough and LONG_CYCLE_RNG is defined, a direct call to LRAN is made.
 * Otherwise, repeated calls to RAND_BASE_CALL are made, concatenating up to 60 random bits.
 * The largest possible multiple of range (with respect to the number of generated bits) is used
 * as a threshold - the generated number is used only if it is less than the threshold, and
 * otherwise rejected and a new number is drawn. This is done to avoid bias towards the smaller
 * values, and so, if RAND_BASE_CALL has a uniform distribution, so does randInt.  				*/
u64 randInt(u64 range){
#if SAFETY_CHECKS
	if(range > RAND_INT_MAX || range == 0){
		LOG_ERR("Error, call to randInt with %" PRIu64 ".\n"
				"randInt must be called with value 0 < k <= %" PRIu64 ".\n", range, RAND_INT_MAX);
		finalize(safety_check_failure);
	}
#endif

// If range is small enough a direct call to LRAN is sufficient.
#ifdef LONG_CYCLE_RNG
	if(range <= INT_MAX){
		return LRAN(range);
	}
#endif

	// Random bits concatenation is required.
	u64 num, threshold = 1;
	int iterations, ii;

#ifdef LONG_CYCLE_RNG
	// Exactly two iterations are required, each generates 30 bits.
	iterations = 2;
	threshold <<= 60;
#endif

#ifdef SIMPLE_RNG
	// Compute how many calls to rand() are required.
	iterations = 1;
	threshold <<= NUM_OF_RAND_BITS;
	while(range > threshold){
		threshold <<= NUM_OF_RAND_BITS;
		iterations += 1;
	}
#endif

	// Compute threshold.
	threshold -= (threshold % range);

	// Keep generating random numbers until a value smaller than threshold is obtained.
	while(true){
		num = RAND_BASE_CALL;
		for(ii = 1; ii < iterations; ii++){
			num = (num << NUM_OF_RAND_BITS) + RAND_BASE_CALL;
		}
		if(num < threshold)
			return (num % range);
	}
}

// Returns the multiplicity (0, 1 or 2) of the edge (u,v).
int countEdgeMultiplicity(int u, int v){
	adjNode *node;
	GET_NODE(u,v,node);
	if(node == NULL)
		return 0;
	if(IS_NODE_OF_SINGLE_EDGE(node))
		return 1;
	return 2;
}

// int-type comparison function given to qsort.
int intCompare(const void *a, const void *b){
	int val1 = *(int*)a;
	int val2 = *(int*)b;

	return val1 - val2;
}

// Comparing function given to qsort when sorting the input degree sequence.
int pairCompare(const void *a, const void *b){
	nd_pair *pair_a = (nd_pair*)a;
	nd_pair *pair_b = (nd_pair*)b;

	return pair_b->d - pair_a->d;
}

// Translates a numeric string to its value if it is in range (quits otherwise).
int strToInt(char* str){
	// Verify string does not begin with a zero.
	if(str[0] == '0'){
		LOG_ERR("ERROR: input parameters must be positive integers with no leading zeroes. Exiting.\n");
		finalize(input_failure);
	}

	// Verify value is in range.
	if(strlen(str) > 10 || atof(str) > S32_MAX){
		LOG_ERR("ERROR: input parameters must not exceed %d. Exiting.\n", S32_MAX);
		finalize(input_failure);
	}

	// Return the numeric value of the string (atoi is safe to use at this point).
	return atoi(str);
}

/* Asks for user's approval to continue when some problem is detected.
 * Quits the program if the user does not approve. 					*/
void approveContinuation(char* message){
	char buffer[100] = "";
	while(strcmp(buffer, "y") && strcmp(buffer,"n")){
		LOG_ERR("%s y/n\n", message);
		scanf("%100s", buffer);
	}
	if(!strcmp(buffer,"n"))
		finalize(user_termination);
}

// Prints the input format for the program.
void printInputFormat(){
	LOG_ERR("Program arguments must be in one of the following formats:\n"
			"1. <input_file>\n"
			"2. <input_file> <output_file>\n"
			"3. <input_file> <output_file> <log_file>\n"
			"or\n"
			"4. <number_of_copies> <n_1> <d_1> <n_2> <d_2> ... <n_k> <d_k>\n\n"

			"If an input file is used, format should match the one in option 4 (all white spaces are permitted).\n\n"

			"number_of_copies graphs will be generated, in each n_i vertices of degree d_i for every i.\n\n"

			"- All parameters must be positive integers with no leading zeroes. \n"
			"- At least one pair n_i d_i must be provided.\n\n"

			"- If an output file is not provided, generated graphs will be printed to stdout.\n"
			"- In order to avoid graphs output, set DO_NOT_OUTPUT to 1 in the code.\n"
			"- If a log file is provided, DEBUG_PRINTS should be set to 1 in the code.\n"
	);
}

// Checks whether the input for the program is via file (true) or via command line (false).
bool checkInputType(int argc, char** argv){
	// Command line input requires an even number > 2 of arguments (including the program name).
	if(!IS_EVEN(argc) || argc == 2)
		return true;

	// Command line input requires each argument to be a numeric string.
	for(int ii = 1; ii < argc; ii++){
		if(NOT_A_NUMBER(argv[ii]))
			return true;
	}

	// Input interpreted as command line input.
	return false;
}

// Extracts data from input file, quits if it is not in the correct format.
char** readInputFile(char* input_file_name, int* input_length){

	// Open the input file.
	FILE *input_file = fopen(input_file_name, "r");
	if(!input_file){
		LOG_ERR("\nCould not open input file \"%s\". Exiting.\n\n", input_file_name);
		printInputFormat();
		finalize(file_failure);
	}

	int  word_count = 0;
	char buffer[12];

	// Verify that every word in the file is a numeric string of length at most 10.
	while(fscanf(input_file, "%11s", buffer) != EOF){
		if(strlen(buffer) == 11 || NOT_A_NUMBER(buffer)){
			fclose(input_file);
			LOG_ERR("\nWrong input file format!\n");
			printInputFormat();
			finalize(input_failure);
		}
		// Count the words for dynamic memory allocation.
		word_count++;
	}

	// Number of arguments must be odd and at least 3.
	if(IS_EVEN(word_count) || word_count == 1){
		fclose(input_file);
		LOG_ERR("\nWrong input file format!\n");
		printInputFormat();
		finalize(input_failure);
	}

	// Input file is in the correct format. Allocate memory and create a word array.
	char *block, **word_array;
	ALLOCATE_MEMORY(block, 		char,  word_count * 11);
	ALLOCATE_MEMORY(word_array, char*, word_count);

	fseek(input_file, 0, SEEK_SET);
	for(int ii = 0; ii < word_count; ii++){
		word_array[ii] = block + 11 * ii;
		fscanf(input_file, "%s", word_array[ii]);
	}

	// Input file was valid and all data was read. Close file and return the word array.
	fclose(input_file);
	*input_length = word_count;
	LOG_DEBUG("Input file was read successfully.\n");
	return word_array;
}

/* Verifies the degree sequence is graphical, using the Erdos-Gallai theorem:
 * A degree sequence d_1 >= d_2 >= ... >= d_n is graphical iff the following hold:
 * 1. The sum of degrees is even;
 * 2. For every 1 <= k < n: Sum_{i <= k} d_i <= k(k-1) + Sum_{i > k} min(d_i, k).
 *
 * We use the improvement of Tripathi-Vijay, saying that the inequality needs to
 * be checked only for the last vertex in each sequence of equal degrees.		*/
void verifyGraphicallity(int length, nd_pair *sequence){

	u64 ii, jj, RHS, sum_of_n_i = 0, sum_of_d_i = 0;

	// Iterate the vertices at the end of each sequence of equal degrees.
	for(ii = 0; ii < length; ii++){
		// Compute the total number of vertices up to this point.
		sum_of_n_i += sequence[ii].n;

		// Compute the LHS.
		sum_of_d_i += sequence[ii].d * sequence[ii].n;

		// Compute the RHS.
		RHS = FALL_2(sum_of_n_i);
		for(jj = ii + 1; jj < length; jj++){
			RHS += fmin(sequence[jj].d, sum_of_n_i) * sequence[jj].n;
		}

		// Check the inequality.
		if(sum_of_d_i > RHS){
			LOG_ERR("\nDegree sequence is not graphical!\n");
			finalize(input_failure);
		}
	}

	// Finally, verify that the sum of degrees is even.
	if(!IS_EVEN(sum_of_d_i)){
		LOG_ERR("\nDegree sequence is not graphical!\n");
		finalize(input_failure);
	}
}
/* Translates the input strings into their numeric values. Verifies that all
 * values are in range and no overflow occurs, and that the degree sequence
 * is graphical (quits otherwise). Computes n, Delta, M and M2. 			*/
nd_pair* verifyInput(char** input, int pairs_count, int* number_of_copies){

	nd_pair *nd_pairs;

	// Set the number of copies to generate.
	*number_of_copies = strToInt(input[0]);

	// Create an array of <n_i, d_i> pairs and verify n, M2 and all d_i are in range.
	ALLOCATE_MEMORY(nd_pairs, nd_pair, pairs_count);
	for(int ii = 0; ii < pairs_count; ii++){
		nd_pairs[ii].n = strToInt(input[2 * ii + 1]);
		nd_pairs[ii].d = strToInt(input[2 * ii + 2]);

		// Update n, M and M2.
		n  += nd_pairs[ii].n;
		M  += nd_pairs[ii].n * nd_pairs[ii].d;
		M2 += nd_pairs[ii].n * FALL_2(nd_pairs[ii].d);

		// Verify the degrees are not too large.
		if(nd_pairs[ii].d > S16_MAX){
			free(nd_pairs);
			LOG_ERR("\nInvalid degree sequence! Degrees must not exceed %d.\n", S16_MAX);
			finalize(input_failure);
		}

		// Verify the number of vertices is not too large.
		if(n < 0){
			free(nd_pairs);
			LOG_ERR("Data overflow. The number of vertices must not exceed %d.\n", S32_MAX);
			finalize(input_failure);
		}

		// Verify M2 is not too large.
		if(M2 > RAND_INT_MAX){
			free(nd_pairs);
			LOG_ERR("Data overflow. M2 must not exceed %" PRIu64 ".\n", RAND_INT_MAX);
			finalize(input_failure);
		}
	}

	// Check for overflow if M_64 is off (i.e. using 32 bits for M).
	if(!M_64 && M > S32_MAX){
		free(nd_pairs);
		LOG_ERR("Data overflow. M must not exceed %d. Set M_64 to 1 in the code.\n", S32_MAX);
		finalize(input_failure);
	}

	// Sort the degrees in descending order and set Delta(G).
	qsort(nd_pairs, pairs_count, sizeof(nd_pair), pairCompare);
	Delta = nd_pairs[0].d;

	// Verify the input is graphical.
	verifyGraphicallity(pairs_count, nd_pairs);

	// Valid input. Return the <n_i,d_i> array.
	LOG_DEBUG("Input is valid.\n");
	return nd_pairs;
}

// If fixed initial seed is used, verifies it has the correct sign according to the RNG.
void checkSeed(){
#if USE_FIXED_SEED
	if(floor(FIXED_SEED) != FIXED_SEED){
		LOG_ERR("\nFIXED_SEED must be an integer. Exiting.\n");
		finalize(configuration_failure);
	}

#ifdef SIMPLE_RNG
	if(FIXED_SEED < 0){
		LOG_ERR("\nSIMPLE_RNG is used with initial seed, but FIXED_SEED is negative (%d). Exiting.\n", FIXED_SEED);
		finalize(configuration_failure);
	}
#endif /* SIMPLE_RNG */

#ifdef LONG_CYCLE_RNG
	if(FIXED_SEED > 0){
		LOG_ERR("\nLONG_CYCLE_RNG is used with initial seed, but FIXED_SEED is positive (%d).\n"
				"Initialization part of the RNG will be skipped.\n\n", FIXED_SEED);
		approveContinuation("Continue without RNG initialization?");
	}
#endif /* LONG_CYCLE_RNG */

#endif /* USE_FIXED_SEED */
}

// Verify there is not any faulty user configuration.
void checkConfiguration(char** argv){
	// Verify that HASH_MULTIPLIER is in range.
	if(HASH_MULTIPLIER <= 0 || HASH_MULTIPLIER > S16_MAX){
		LOG_ERR("ERROR! HASH_MULTIPLIER must be a positive constant at most %d.\n", S16_MAX);
		finalize(configuration_failure);
	}

	// If fixed initial seed is used, verify it has the correct sign according to the RNG.
	checkSeed();

	// Check output configuration consistency.
	if(DO_NOT_OUTPUT && output_file != stdout){
		LOG_ERR("\nAn output file (%s) was provided but graph output is turned off.\n"
				"In order to enable graph output set DO_NOT_OUTPUT to 0 in the code.\n\n"
				"Proceeding without graph output.\n\n", argv[2]);
	}
}

// Initializes the RNG according to its definitions.
void initializeRNG(){
#ifdef LONG_CYCLE_RNG
#if USE_FIXED_SEED
	seed = FIXED_SEED;
#else
	// Use random initial seed.
	srand(time(NULL));
	seed = -rand();
#endif /* USE_FIXED_SEED */
#endif /* LONG_CYCLE_RNG */

#ifdef SIMPLE_RNG
#if USE_FIXED_SEED
	srand(FIXED_SEED);
#else
	// Use random initial seed.
	srand(time(NULL));
#endif /* USE_FIXED_SEED */
#endif /* SIMPLE_RNG */
}

// Allocates all dynamic memory and initializes all the data that depends on the degree sequence alone.
void initialize(nd_pair *nd_pairs, int num_of_pairs){
	int 	ii, jj;
	int 	curr_vertex = 0, curr_hash_cells_num, curr_hash_cells_size;
	u64  	sum_deg = 0;
	adjNode **hash_pos;

	// Initialize the random number generator.
	initializeRNG();

	// Compute B1 and B2.
	if(22 * Delta * Delta * Delta < M2){
		B1 = M2 / M;
		B2 = floor((1.0 * M2 * M2) / (1.0 * M * M));
	}// else B1 = B2 = 0.

	// Compute the total number of hash cells required.
	for(ii = 0; ii < num_of_pairs; ii++){
		total_hash_cells += ceil(nd_pairs[ii].d * (HASH_MULTIPLIER)) * nd_pairs[ii].n;
	}

	// Allocate all dynamic memory.
	ALLOCATE_MEMORY(V, 			  vertex,   n);
	ALLOCATE_MEMORY(points, 	  int,      M);
	ALLOCATE_MEMORY(adjNode_pool, adjNode,  M);
	ALLOCATE_MEMORY(hash_array,   adjNode*, total_hash_cells);
	ALLOCATE_MEMORY(encountered,  int, 	    Delta * Delta);

	// C6_array is the only one needed to be set to 0.
	ALLOCATE_INITIALIZED_MEMORY(C6_array, int, n);

	// Create the points array and initialize the vertex structure of each v_i.
	for(hash_pos = hash_array, ii = 0; ii < num_of_pairs; ii++){
		/* Compute the number of hash cells to be allocated to each vertex of
		 * the current degree, and how many vertices are mapped to each cell. */
		curr_hash_cells_num  = ceil(nd_pairs[ii].d * 1.0 * (HASH_MULTIPLIER));
		curr_hash_cells_size = ceil((1.0 * n) / curr_hash_cells_num);

		// Handle all vertices with the current degree.
		while(nd_pairs[ii].n > 0){
			// "single_edges_num" and "next_available_node" are initialized with each pairing generation.
			V[curr_vertex].has_loop		  = false;
			V[curr_vertex].degree		  = nd_pairs[ii].d;
			V[curr_vertex].hash_cell_size = curr_hash_cells_size;
			V[curr_vertex].neighbors_pool = adjNode_pool + sum_deg;
			V[curr_vertex].hash			  = hash_pos;

			// Points corresponding to the current vertex.
			for(jj = 0; jj < nd_pairs[ii].d; jj++){
				points[sum_deg] = curr_vertex;
				sum_deg++;
			}

			// Update counters.
			hash_pos += curr_hash_cells_num;
			curr_vertex++;
			nd_pairs[ii].n--;
		}
	}

	free(nd_pairs);
	LOG_DEBUG("\nInitialization completed successfully.\n"
			  "n = %d, Delta = %" PRIu64 ", M = %" PRIu64 ", M2 = %"PRIu64", B1 = %d, B2 = %d.\n",
			  n, Delta, M, M2, B1, B2);
}

// Prints generated pairing related statistics at the end of the run.
void printStatisticsPairings(int generation_fails, int total_attempts){
	if(generation_fails == 0){
		LOG_INFO("\n----- All pairing generations were successful.\n");
		return;
	}

	LOG_INFO("\nPairing generations:\n"
			 "---------------------\n"
			 "%d successful, %d failed. Failures according to the first detected problem:\n"
			 "double loop %d, triple edge %d, too many loops %d, too many double edges %d.\n",
			 total_attempts - generation_fails, generation_fails,
			 double_loop, triple_edge, many_loops, many_double_edges);
}

// Print loops related statistics at the end of the run.
void printStatisticsLoops(){
	if(total_L_switch_calls == 0){
		if(B1 == 0)
			LOG_INFO("\n----- No loops were allowed.\n");
		else
			LOG_INFO("\n----- No calls to NoLoops were made.\n");

		return;
	}

	int total_rejections = loop_fwd_rej + loop_bck_rej;
	int success = total_L_switch_calls - total_rejections;
	LOG_INFO("\nLoops:\n"
			 "-------\n"
			 "Total calls to NoLoops %d, average loops per call %.1f, total rejections %d (%.1f%%).\n"
			 "Total loop removal attempts %d, Forward rejections %d (%.1f%%), Backward rejections %d (%.1f%%), "
			 "success %d (%.1f%%).\n",
			 total_noloops, 1.0 * total_loop_num / total_noloops, VAL_AND_PERCENT(total_rejections ,total_noloops),
			 total_L_switch_calls, VAL_AND_PERCENT(loop_fwd_rej ,total_L_switch_calls),
			 VAL_AND_PERCENT(loop_bck_rej, total_L_switch_calls), VAL_AND_PERCENT(success, total_L_switch_calls));
}

// Print double edges related statistics at the end of the run.
void printStatisticsDoubles(){
	if(total_D_switch_calls == 0){
		if(B2 == 0)
			LOG_INFO("\n----- No double edges were allowed.\n");
		else
			LOG_INFO("\n----- No calls to NoDoubles were made.\n");

		return;
	}

	int total_rejections, success;

	total_rejections = double_fwd_rej + double_bck_rej;
	success 		 = total_D_switch_calls - total_rejections;
	LOG_INFO("\nDouble edges:\n"
			 "--------------\n"
			 "Total calls to NoDoubles %d, average double edges per call %.1f, total rejections %d (%.1f%%).\n"
			 "Total double edge removal attempts %d, Forward rejections %d (%.1f%%), "
			 "Backward rejections %d (%.1f%%), success %d (%.1f%%).\n",
			 total_nodoubles, 1.0 * total_double_edges_num / total_nodoubles,
			 VAL_AND_PERCENT(total_rejections ,total_nodoubles),
			 total_D_switch_calls, VAL_AND_PERCENT(double_fwd_rej ,total_D_switch_calls),
			 VAL_AND_PERCENT(double_bck_rej, total_D_switch_calls), VAL_AND_PERCENT(success, total_D_switch_calls));
}

/* Prints all pairs in the points array according to their multiplicity
 * and by their order: loops, single edges, double edges. 			 */
void _printPairing(){
	for(u64 ii = 0; ii < M/2; ii++)
		LOG_DEBUG("pair %" PRIu64": %d - %d\n", ii, points[2 * ii], points[2 * ii + 1]);
}

// Prints the "edge" field for all allocated adjacency nodes, vertex by vertex.
void printEdgeLocations(){
	adjNode *node;
	for(int ii = 0; ii < n; ii++){
		for(node = START_OF_POOL(ii); node < END_OF_POOL(ii) - 1; node++){
			LOG_DEBUG("(%d,%d) --> %" PRINT_M", ", ii, node->id, node->edge);
		}
		LOG_DEBUG("(%d,%d) --> %" PRINT_M".\n", ii, node->id, node->edge);
	}
}

// Prints all neighbors of a given vertex by iterating its neighbors pool.
void _printNeighborsByPool(int u){
	adjNode *node;
	LOG_DEBUG("Vertex %d has degree %d. Neighbors: ",u, V[u].degree);
	for(node = START_OF_POOL(u); node < END_OF_POOL(u) - 1; node++)
		LOG_DEBUG("%d, ", node->id);
	LOG_DEBUG("%d", node->id);
	if(V[u].has_loop)
		LOG_DEBUG(", and also has a loop.\n");
	else
		LOG_DEBUG(".\n");
}

// Prints all neighbors of a given vertex by iterating its hash.
void _printNeighborsByHash(int u){
	adjNode* node;
	LOG_DEBUG("Vertex %d has degree %d (cell size = %d, single edges = %d). Neighbors: ",
				u, V[u].degree, V[u].hash_cell_size, V[u].single_edges_num);

	for(int ii = 0; ii < ceil(V[u].degree * 1.0 * (HASH_MULTIPLIER)); ii++){
		node = V[u].hash[ii];
		LOG_DEBUG("[");
		if(node == NULL) LOG_DEBUG("*");
		while(node != NULL){
			LOG_DEBUG("%d", node->id);
			if(node->edge == M)
				LOG_DEBUG("(D)");
			node = node->next;
			if(node != NULL)
				LOG_DEBUG(",");
		}
		LOG_DEBUG("]");
	}
	if(V[u].has_loop)
		LOG_DEBUG(" and also has a loop.\n");
	else
		LOG_DEBUG(".\n");
}

// Iterates all vertices and for each prints all its neighbors.
void _printAllNeighbors(){
/*
	LOG_DEBUG("Print by pool:\n");
	for(int ii = 0; ii < n; ii++)
		printNeighborsByPool(ii);
*/

	LOG_DEBUG("Print by hash:\n");
	for(int ii = 0; ii < n; ii++)
		_printNeighborsByHash(ii);
}

// Outputs all edges in the graph by their order in the points array.
void outputGraphUnsortedEdges(){
	char buffer[100];

	// Set the column width.
	int width = COLUMN_WIDTH;
	if(width == 0)
		width = log(n) + 8;

	for(u64 ii = 0; ii < M/2; ii++){
		if(ii % EDGES_PER_LINE == 0)
			OUTPUT("\n");
		sprintf(buffer, "(%d,%d)", points[2 * ii], points[2 * ii + 1]);
		OUTPUT("%-*s", width, buffer);
	}
	OUTPUT("\n\n\n");
}

// Outputs the graph as unordered lists of neighbors per vertex.
void outputGraphUnsortedNeighbors(){
	adjNode *node;

	// Iterate all vertices.
	for(int ii = 0; ii < n; ii++){
#if PRINT_EMPTY_LISTS
		OUTPUT("\n%d: ", ii);
		node = START_OF_POOL(ii);
#else
		// Look for the first neighbor.
		for(node = START_OF_POOL(ii); node < END_OF_POOL(ii); node++){
			if(node->id > ii){
				OUTPUT("\n%d: %d ", ii, node->id);
				node++;
				break;
			}
		}
#endif

		// Iterate remaining/all neighbors of the current vertex.
		for(; node < END_OF_POOL(ii); node++){
			if(node->id > ii){
				OUTPUT("%d ", node->id);
			}
		}
	}

	OUTPUT("\n\n\n");
}

/* Outputs the graph in a sorted way, according to user configuration.
 * If OUTPUT_GRAPH_BY_VERTEX is on, prints for each vertex its ordered
 * 									list of neighbors of larger index.
 * Otherwise, prints all edges in the graph in lexicographical order. */
void outputGraphSorted(){
	int 	ii, jj, cell_length = 0, vertex_length = 0;
	int 	*list_to_array;
	adjNode *node, **head;

	ALLOCATE_MEMORY(list_to_array, int, Delta);

#if !OUTPUT_GRAPH_BY_VERTEX
	int  total_edges_printed = 0;
	char buffer[100];

	// Set the column width.
	int width = COLUMN_WIDTH;
	if(width == 0)
		width = log(n) + 8;
#endif

	// Iterate all vertices.
	for(ii = 0; ii < n - 1; ii++){
		/* For each vertex, iterate its hash table. There is order between cells, but not within each cell.
		 * Skip all cells corresponding to vertices of smaller index than ii. */
		for(head = GET_CELL(ii, ii + 1); head < V[ii + 1].hash; head++){
			// Sort every non-empty hash cell.
			if(*head != NULL){
				node = *head;
				while(node != NULL){
					list_to_array[vertex_length + cell_length] = node->id;
					cell_length++;
					node = node->next;
				}

				qsort(list_to_array + vertex_length, cell_length, sizeof(int), intCompare);
				vertex_length += cell_length;
				cell_length = 0;
			}
		}

		jj = 0;
		// Skip neighbors with index smaller than ii (may appear only in the first iterated cell of the vertex).
		while((jj < vertex_length) && (list_to_array[jj] < ii))
			jj++;

#if OUTPUT_GRAPH_BY_VERTEX
		if(jj < vertex_length || PRINT_EMPTY_LISTS)
			OUTPUT("\n%d: ", ii);
		while(jj < vertex_length){
			OUTPUT("%d ", list_to_array[jj]);
			jj++;
		}
#else
		// Print all other edges.
		while(jj < vertex_length){
			if(total_edges_printed % EDGES_PER_LINE == 0)
				OUTPUT("\n");
			total_edges_printed++;
			sprintf(buffer, "(%d,%d)", ii, list_to_array[jj]);
			OUTPUT("%-*s", width, buffer);
			jj++;
		}
#endif

		vertex_length = 0;
	}

#if OUTPUT_GRAPH_BY_VERTEX
	// The last vertex has no neighbors of larger index.
	if(PRINT_EMPTY_LISTS)
		OUTPUT("\n%d:", n-1);
#endif

	free(list_to_array);
	OUTPUT("\n\n\n");
}

// Verifies each vertex appears the correct number of times in the points array (quits otherwise).
void verifyPointsCount(){
	u64 ii;
	int *count;

	ALLOCATE_INITIALIZED_MEMORY(count, int, n);

	for(ii = 0; ii < M; ii++)
		count[points[ii]]++;

	for(ii = 0; ii < n; ii++){
		if(V[ii].degree != count[ii]){
			LOG_ERR("ERROR: vertex %" PRIu64 " has degree %d but appears %d times.\n",
					ii, V[ii].degree, count[ii]);
			_printPairing();
			free(count);
			finalize(safety_check_failure);
		}
	}

	free(count);
	LOG_DEBUG("pairing count verified.\n");
}

// Verifies the pairs in the non-single sections are valid (quits otherwise).
void verifyPointsSections(){
	u64 ii;

	// Check loops section.
	for(ii = 0; ii < START_OF_SINGLE_EDGES; ii += 2){
		if(points[ii] != points[ii + 1]){
			LOG_ERR("ERROR: pair %" PRIu64 " should be a loop but is not.\n", ii / 2);
			_printPairing();
			finalize(safety_check_failure);
		}
	}

	// Check double edges section.
	for(ii = START_OF_DOUBLE_EDGES; ii < M; ii += 4){
		if(!ARE_PAIRS_EQUAL(points[ii], points[ii + 1], points[ii + 2], points[ii + 3])){
			LOG_ERR("ERROR: pairs %" PRIu64 " - %" PRIu64
					" should represent a double edge but are not.\n", ii / 2, ii / 2 + 1);
			_printPairing();
			finalize(safety_check_failure);
		}
	}

	LOG_DEBUG("pairing sections verified.\n");
}

// Verifies the points array is proper (quits otherwise).
void _checkPointsArray(){
	// Verify each vertex appears the correct number of times in the points array.
	verifyPointsCount();

	// Verify the pairs in the non-single sections are valid.
	verifyPointsSections();
}

// Verifies the correctness of the "edge" field of every adjacency node (quits otherwise).
void verifyEdgeIndicators(){
	u64 	jj;
	adjNode *node;

	for(int ii = 0; ii < n; ii++){
		for(node = START_OF_POOL(ii); node < END_OF_POOL(ii); node++){
			// Verify "edge" field is in range.
			if((node->edge < 0) || (node->edge > M)){
				LOG_ERR("ERROR: edge %d->%d indicator wrong (marked %" PRINT_M").\n",
						ii, node->id, node->edge);
				finalize(safety_check_failure);
			}

			if(IS_NODE_OF_DOUBLE_EDGE(node)){
				// Verify the edge appears in the double edges section in the points array.
				for(jj = START_OF_DOUBLE_EDGES; jj < M; jj += 4){
					if(ARE_PAIRS_EQUAL(ii, node->id, points[jj], points[jj + 1]))
						break;
				}
				if(jj >= M){
					LOG_ERR("ERROR: edge %d->%d indicator wrong (marked double).\n", ii, node->id);
					_printNeighborsByHash(ii);
					printEdgeLocations();
					finalize(safety_check_failure);
				}
			}
		}
	}
	LOG_DEBUG("Edge indicators verified.\n");
}

// Verifies the the correctness and consistency of the different data structures (quits otherwise).
void _verifyDataStructures(){

	u64		ii;
	adjNode *node;
	bool 	faulty_database = false;
	int 	singles, doubles;
	int 	*double_edge_count;

	// Verifies the points array is proper.
	_checkPointsArray();

	// Verify the correctness of the "edge" field of every adjacency node.
	verifyEdgeIndicators();

	// Verify for each vertex that the loops, single edges, and double edges incident with it sum up to its degree.
	ALLOCATE_INITIALIZED_MEMORY(double_edge_count, int, n);
	for(ii = START_OF_DOUBLE_EDGES; ii < M; ii++){
		double_edge_count[points[ii]]++;
	}
	for(ii = 0; ii < n; ii++){
		if(V[ii].degree != V[ii].single_edges_num + 2 * V[ii].has_loop + double_edge_count[ii]){
			LOG_ERR("ERROR: vertex %" PRIu64 " has degree %d but single edges number is %d, "
					"double edges count is %d, and it has %s loop.\n",
					ii, V[ii].degree, V[ii].single_edges_num, double_edge_count[ii],
					V[ii].has_loop? "a" : "no");
			_printNeighborsByHash(ii);
			faulty_database = true;
		}

		singles = doubles = 0;
		for(node = START_OF_POOL(ii); node < END_OF_POOL(ii); node++){
			if(node->edge < M)
				singles++;
			else
				doubles++;
		}
		if(singles != V[ii].single_edges_num){
			LOG_ERR("ERROR: vertex %" PRIu64 " has %d single nodes but single edges number is %d.\n",
					ii, singles, V[ii].single_edges_num);
			_printNeighborsByHash(ii);
			faulty_database = true;
		}
		if(2 * doubles != double_edge_count[ii]){
			LOG_ERR("ERROR: vertex %" PRIu64 " has %d double nodes but double edges count is %d.\n",
					ii, doubles, double_edge_count[ii]);
			_printNeighborsByHash(ii);
			faulty_database = true;
		}
	}

	free(double_edge_count);

	if(faulty_database){
		finalize(safety_check_failure);
	}

	// Verify that P2 holds the correct value.
	u64 tmp_P2 = 0;
	for(ii = 0; ii < n; ii++){
		if(!V[ii].has_loop)
			tmp_P2 += FALL_2(V[ii].single_edges_num);
	}

	if(P2 != tmp_P2){
		LOG_ERR("ERROR: P2 %"PRIu64" != tmp_P2 %"PRIu64".\n", P2, tmp_P2);
		printAllNeighbors();
		finalize(safety_check_failure);
	}

	LOG_DEBUG("Data structures verified.\n");
}

// Adds a new single neighbor to a vertex u.
void addNewSingleNeighbor(int u, int neighbor, type_M edge_location){
#if SAFETY_CHECKS
	if(END_OF_POOL(u) == START_OF_POOL(u) + V[u].degree){
		LOG_ERR("ERROR: vertex %d has already allocated all its nodes!\n", u);
		finalize(safety_check_failure);
	}
	if(IS_EDGE(u, neighbor)){
		LOG_ERR("ERROR: vertex %d is already a neighbor of %d!\n", neighbor, u);
		finalize(safety_check_failure);
	}
#endif

	// Allocate a new adjacency node.
	adjNode *node = END_OF_POOL(u);
	END_OF_POOL(u)++;

	// Update the node's fields.
	node->id   = neighbor;
	node->edge = edge_location;

	// Insert node to hash.
	HEAD_INSERT(u, node);
}

// Changes one of the neighbors of a vertex u to a new, single, neighbor.
void changeNeighbor(int u, int old_neighbor, int new_neighbor){
	LOG_DEBUG("Adding %d as a neighbor of %d instead of %d.\n", new_neighbor, u, old_neighbor);

#if SAFETY_CHECKS
	if(IS_NO_EDGE(u, old_neighbor)){
		LOG_ERR("ERROR: vertex %d is not a neighbor of %d!\n", old_neighbor, u);
		finalize(safety_check_failure);
	}
	if(IS_EDGE(u, new_neighbor)){
		LOG_ERR("ERROR: vertex %d is already a neighbor of %d!\n", new_neighbor, u);
		finalize(safety_check_failure);
	}
#endif

	adjNode* node;

	// If the two vertices belong to the same hash cell, just update the content of the node.
	if(GET_CELL(u, old_neighbor) == GET_CELL(u, new_neighbor)){
		GET_NODE(u, old_neighbor, node);
		node->id   = new_neighbor;
		node->edge = 0;
		return;
	}

	// The vertices belong to different hash cells. First extract node from old cell.
	EXTRACT_NODE(u, old_neighbor, node);

	// Update node's fields.
	node->id   = new_neighbor;
	node->edge = 0;

	// Insert the new neighbor at the top of its hash cell.
	HEAD_INSERT(u, node);
}

/* Swaps the location in the points array of a single edge and a newly detected double edge.
 * Used when a new pairing is generated and double edges are compressed to the end of the points array. */
void edgeLocationSwap(u64 single_edge_new_location, u64 double_edge_new_location){
	adjNode* node;

	LOG_DEBUG("Swapping pair %" PRIu64": %d - %d with pair %" PRIu64": %d - %d.\n",
				single_edge_new_location / 2,
				points[single_edge_new_location], points[single_edge_new_location + 1],
				double_edge_new_location / 2,
				points[double_edge_new_location], points[double_edge_new_location + 1]);

	// Update the points array.
	POINTS_SWAP(single_edge_new_location, double_edge_new_location);
	POINTS_SWAP(single_edge_new_location + 1, double_edge_new_location + 1);

	// Update edge location in case the single edge later turns out to be a double edge.
	GET_NODE(points[single_edge_new_location], points[single_edge_new_location + 1], node);
	node->edge =  single_edge_new_location;
	GET_NODE(points[single_edge_new_location + 1], points[single_edge_new_location], node);
	node->edge =  single_edge_new_location;
}

// Reject (i.e. return true) with probability 1 - (num1*num2)/(denom1*denom2).
bool checkBackwardsRejection(u64 num1, u64 num2, u64 denom1, u64 denom2){
	// Generate a random number just once if possible.
	if(RAND_INT_MAX / denom1 >= denom2){
		return(randInt(denom1 * denom2) >= num1 * num2);
	}
	else{
		// Compute incrementally.
		return(randInt(denom1) >= num1 || randInt(denom2) >= num2);
	}
}

/* Generates a random pairing of all the points and creates all relevant data structures accordingly.
 * All pairs corresponding to loops are compressed to the beginning of the points array, and all pairs corresponding
 * to double edges are compressed to the end of the points array (each double edge appears twice consecutively).
 * Returns false if one of the following occurs:
 * 		a double loop; a triple edge; too many loops; too many double edges.
 * Otherwise, returns true.  																					 */
bool generatePairing(){

	int 	val1, val2;
	type_M 	ii, randNum;
	adjNode	*existing_node;

	// First initialize all relevant data structures.
	double_count = 0;
	memset(hash_array, 0, sizeof(adjNode*) * total_hash_cells);
	for(ii = 0; ii < n; ii++){
		V[ii].single_edges_num = 0;
		END_OF_POOL(ii)	= START_OF_POOL(ii);
	}

	// loop_count and the points array still hold the information from the previous iteration.
	while(loop_count > 0){
		loop_count--;
		V[points[START_OF_SINGLE_EDGES]].has_loop = false;
	}

	// Generate the pairing.
	for(ii = M - 1; ii > 1; ii--){
		randNum = randInt(ii);
		ii--;
		POINTS_SWAP(ii, randNum);
	}

	printPairing();
	checkPointsArray();

	// Sort points array: loops - single edges - double edges. Add adjacency nodes for each pair.
	for(ii = M - 2; ii >= START_OF_SINGLE_EDGES; ii -= 2){
		val1 = points[ii];
		val2 = points[ii + 1];
		if(val1 == val2){

			// A loop has been detected
			if(V[val1].has_loop){
				// Double loop at v_i. The pairing is invalid.
				LOG_DEBUG("double loop at %d.\n", val1);
				double_loop++;
				return false;
			}

			// This is a single loop. Check if the limit on loops is not exceeded.
			if(loop_count == B1){
				// Too many loops. The pairing is invalid.
				LOG_DEBUG("too many loops (limit is %d).\n", B1);
				many_loops++;
				return false;
			}

			// Compress all loops to the beginning of the points array.
			POINTS_SWAP(ii, START_OF_SINGLE_EDGES);
			POINTS_SWAP(ii + 1, START_OF_SINGLE_EDGES + 1);

			// Update loop count and data for the vertex.
			loop_count++;
			V[val1].has_loop = true;

			// Increase counter to handle the swapped pair in the next iteration.
			ii += 2;
		}
		else{
			// A non-loop is detected. Check whether v2 is already a neighbor of v1.
			GET_NODE(val1, val2, existing_node);
			if(existing_node == NULL){
				// First appearance of this edge. Create new single adjacency nodes for both vertices.
				addNewSingleNeighbor(val1, val2, ii);
				addNewSingleNeighbor(val2, val1, ii);

				// Update single edges count for both vertices.
				V[val1].single_edges_num++;
				V[val2].single_edges_num++;
			}
			else{
				// Multiple edge detected.
				if(IS_NODE_OF_DOUBLE_EDGE(existing_node)){
					// Triple edge. The pairing is invalid.
					LOG_DEBUG("triple edge at %d-%d.\n", val1, val2);
					triple_edge++;
					return false;
				}

				// This is a double edge. Check if the limit on double edges is not exceeded.
				if(double_count == B2){
					// Too many double edges. The pairing is invalid.
					LOG_DEBUG("too many double edges (limit is %d).\n", B2);
					many_double_edges++;
					return false;
				}
				// This is a valid double edge.
				double_count++;

				// Compress double edges to the end of the points array.
				edgeLocationSwap(existing_node->edge, START_OF_DOUBLE_EDGES + 2);
				edgeLocationSwap(ii, START_OF_DOUBLE_EDGES);

				// Mark both adjacency nodes as representing a double edge.
				existing_node->edge = M;
				GET_NODE(val2, val1, existing_node);
				existing_node->edge = M;

				// Update single edges count for both vertices.
				V[val1].single_edges_num--;
				V[val2].single_edges_num--;
			}
		}
	}

	// Compute P2.
	P2 = 0;
	for(ii = 0; ii < n; ii++){
		if(!V[ii].has_loop)
			P2 += FALL_2(V[ii].single_edges_num);
	}

	printPairing();
	printAllNeighbors();
	verifyDataStructures();

	return true;
}

/* Attempts to perform an L-switch. Labeling according to the paper.
 * A random loop v2 and two random single edges (v1,v4) and (v3,v5) are
 * replaced (if possible) by the edges (v1,v2), (v2,v3) and (v4,v5).
 * In case of vertex collision or forbidden edges returns false (f-rejection).
 * Otherwise, returns true and updates S to be the 2-path v1-v2-v3.			*/
bool performL(int* S){

	int v1, v2, v3, v4, v5;
	u64 loop, single_edge1, single_edge2;

	// Selects a random loop and two random single edges.
	loop  		 = RANDOM_LOOP;
	single_edge1 = RANDOM_SINGLE_EDGE;
	single_edge2 = RANDOM_SINGLE_EDGE;


	/* Label all participating vertices and point edges to their left cells
	 * in the points array (the loop already points to its left cell).	 */
	v2 = points[loop];
	ADJUST_EDGE(v1, v4, single_edge1);
	ADJUST_EDGE(v3, v5, single_edge2);

	LOG_DEBUG("Selection for L-Switch. Loop: %d (pair %"PRIu64"), edge 1: %d-%d (pair %"PRIu64"), "
			  "edge 2: %d-%d (pair %"PRIu64").\n", v2, loop/2, v1, v4, single_edge1/2, v3, v5, single_edge2/2);

	// Check for vertex collision.
	if(v2 == v1 || v2 == v3 || v2 == v4 || v2 == v5 || v1 == v3 || v1 == v5 || v4 == v3 || v4 == v5){
		LOG_DEBUG("Vertex collision while attempting to perform L-Switch.\n");
		return false;
	}

	// Check for forbidden edges.
	if(IS_EDGE(v1,v2) || IS_EDGE(v2,v3) || IS_EDGE(v4,v5)){
		LOG_DEBUG("Forbidden edge(s) while attempting to perform L-Switch:%s%s%s.\n",
					IS_EDGE(v1,v2)? " v1-v2" : "", IS_EDGE(v2,v3)? " v2-v3" : "", IS_EDGE(v4,v5)? " v4-v5" : "");

		printNeighborsByHash(v1);
		printNeighborsByHash(v2);
		printNeighborsByHash(v3);
		printNeighborsByHash(v4);
		printNeighborsByHash(v5);
		return false;
	}

	// L-Switch can be performed. Update S.
	S[0] = v1;
	S[1] = v2;
	S[2] = v3;

	// Compress loops to the beginning of the points array.
	loop_count--;
	points[loop]	 = points[START_OF_SINGLE_EDGES];
	points[loop + 1] = points[START_OF_SINGLE_EDGES + 1];

	// Update new edges in the points array.
	points[single_edge1] 			  = v1;
	points[single_edge1 + 1] 		  = v2;
	points[single_edge2] 			  = v2;
	points[single_edge2 + 1] 		  = v3;
	points[START_OF_SINGLE_EDGES] 	  = v4;
	points[START_OF_SINGLE_EDGES + 1] = v5;

	// Mark v2 as not having a loop anymore and add its contribution to P2.
	V[v2].has_loop = false;
	V[v2].single_edges_num += 2;
	P2 += FALL_2(V[v2].single_edges_num);

	// Add v1 and v3 as new single neighbors of v2.
	addNewSingleNeighbor(v2, v1, 0);
	addNewSingleNeighbor(v2, v3, 0);

	// For the other vertices change adjacency nodes.
	changeNeighbor(v1, v4, v2);
	changeNeighbor(v4, v1, v5);
	changeNeighbor(v3, v5, v2);
	changeNeighbor(v5, v3, v4);

	LOG_DEBUG("S: [%d,%d,%d], e: (%d,%d).\n", v1, v2, v3, v4, v5);
	printPairing();
	printNeighborsByHash(v1);
	printNeighborsByHash(v2);
	printNeighborsByHash(v3);
	printNeighborsByHash(v4);
	printNeighborsByHash(v5);

	return true;
}

/* Attempts to perform a D-switch. Labeling according to the paper.
 * A random double edge (v2,v5) and two random single edges (v1,v4) and (v3,v6)
 * are replaced (if possible) by the edges (v1,v2), (v2,v3), (v4,v5) and (v5,v6).
 * In case of vertex collision or forbidden edges returns false (f-rejection).
 * Otherwise, returns true and updates S to be the 2-path v1-v2-v3.				*/
bool performD(int* S){

	int v1, v2, v3, v4, v5, v6;
	u64 double_edge, single_edge1, single_edge2;

	// Select a random double edge and two random single edges.
	double_edge  = RANDOM_DOUBLE_EDGE;
	single_edge1 = RANDOM_SINGLE_EDGE;
	single_edge2 = RANDOM_SINGLE_EDGE;

	// Point the double edge to its leftmost copy in the points array while maintaining orientation.
	double_edge -= (((double_edge - START_OF_DOUBLE_EDGES) % 4) / 2) * 2;

	// Label all participating vertices and point edges to their left cells in the points array.
	ADJUST_EDGE(v2, v5, double_edge);
	ADJUST_EDGE(v1, v4, single_edge1);
	ADJUST_EDGE(v3, v6, single_edge2);

	LOG_DEBUG("Selection for D-Switch. Double edge: %d-%d (pair %"PRIu64"), "
			  "single edge 1: %d-%d (pair %"PRIu64"), single edge 2: %d-%d (pair %"PRIu64").\n",
			  v2, v5, double_edge/2, v1, v4, single_edge1/2, v3, v6, single_edge2/2);

	// Check for vertex collision.
	if(v2 == v1 || v2 == v3 || v2 == v4 || v2 == v6 ||
	   v5 == v1 || v5 == v3 || v5 == v4 || v5 == v6 ||
	   v1 == v3 || v1 == v6 || v4 == v3 || v4 == v6){
		LOG_DEBUG("Vertex collision while attempting to perform D-Switch.\n");
		return false;
	}

	// Check for forbidden edges.
	if(IS_EDGE(v1,v2) || IS_EDGE(v2,v3) || IS_EDGE(v4,v5) || IS_EDGE(v5,v6)){
		LOG_DEBUG("Forbidden edge(s) while attempting to perform D-Switch:%s%s%s%s.\n",
					IS_EDGE(v1,v2)? " v1-v2" : "", IS_EDGE(v2,v3)? " v2-v3" : "",
					IS_EDGE(v4,v5)? " v4-v5" : "", IS_EDGE(v5,v6)? " v5-v6" : "");
		printNeighborsByHash(v1);
		printNeighborsByHash(v2);
		printNeighborsByHash(v3);
		printNeighborsByHash(v4);
		printNeighborsByHash(v5);
		printNeighborsByHash(v6);
		return false;
	}

	// D-Switch can be performed. Update S.
	S[0] = v1;
	S[1] = v2;
	S[2] = v3;

	// Compress double edges to the end of the points array.
	for(int ii = 0; ii < 4; ii++){
		points[double_edge + ii] = points[START_OF_DOUBLE_EDGES + ii];
	}

	// Update new edges in the points array.
	points[START_OF_DOUBLE_EDGES]		= v1;
	points[START_OF_DOUBLE_EDGES + 1]	= v2;
	points[START_OF_DOUBLE_EDGES + 2] 	= v2;
	points[START_OF_DOUBLE_EDGES + 3] 	= v3;
	points[single_edge1] 				= v4;
	points[single_edge1 + 1] 			= v5;
	points[single_edge2] 				= v5;
	points[single_edge2 + 1] 			= v6;
	double_count--;

	// Update / create relevant adjacency nodes.
	// v2 and v5 have an unused node in their nodes pool due to the double edge.
	addNewSingleNeighbor(v2, v1, 0);
	addNewSingleNeighbor(v5, v4, 0);

	// For all other new single edges use existing nodes.
	changeNeighbor(v2, v5, v3);
	changeNeighbor(v5, v2, v6);
	changeNeighbor(v1, v4, v2);
	changeNeighbor(v4, v1, v5);
	changeNeighbor(v3, v6, v2);
	changeNeighbor(v6, v3, v5);

	// If v2/v5 was incident to r single edges its contribution to P2 is increased by (r+2)(r+1) - r(r-1) = 4r + 2.
	P2 += (4 * V[v2].single_edges_num + 2) + (4 * V[v5].single_edges_num + 2);
	V[v2].single_edges_num += 2;
	V[v5].single_edges_num += 2;

	LOG_DEBUG("S: [%d,%d,%d], P: %d-%d-%d.\n", v1, v2, v3, v4, v5, v6);
	printPairing();
	printNeighborsByHash(v1);
	printNeighborsByHash(v2);
	printNeighborsByHash(v3);
	printNeighborsByHash(v4);
	printNeighborsByHash(v5);
	printNeighborsByHash(v6);

	return true;
}

/* For a given simple 2-path S = v1-v2-v3, returns the number of ordered single edges e =(v4,v5) such that
 * e is vertex disjoint from S, and (v1,v4) and (v2,v5) are non-edges (labeling according to the paper).
 * Done by counting the "bad" pairs and subtracting them from M, the total number of ordered pairs.		*/
u64 computeBLG(int* S){
	int		v1 = S[0], v2 = S[1], v3 = S[2], v4, v5;
	u64		ii, count;
	adjNode *node, *internal_node;

	// First count unordered pairs which do not correspond to simple edges.
	count = loop_count + 2 * double_count;

	// Second, count unordered simple edges intersecting S.
	for(ii = 0; ii < 3; ii++){
		count += V[S[ii]].single_edges_num;
	}
	// Subtract simple edges with both endpoints in S which were double counted.
	if(IS_SINGLE_EDGE(v1, v3))
		count -= 3;
	else
		count -= 2;

	// The count so far was of unoriented edges, proceed with oriented edges.
	count *= 2;

	// Now count ordered simple edges (v4,v5) disjoint from S, such that a forbidden edge exists.

	// Iterate all neighbors v4 of v1 not in S.
	for(node = START_OF_POOL(v1); node < END_OF_POOL(v1); node++){
		v4 = node->id;
		if(v4 != v2 && v4 != v3){
			// Add to the count all single edges (v4,v5) such that v5 is not in S.
			count += V[v4].single_edges_num;
			for(ii = 0; ii < 3; ii++){
				if(IS_SINGLE_EDGE(v4, S[ii]))
					count--;
			}
		}
	}

	// Now iterate all neighbors v5 of v3.
	for(node = START_OF_POOL(v3); node < END_OF_POOL(v3); node++){
		v5 = node->id;
		if(v5 != v1 && v5 != v2){
			/* Instead of just counting the single edges incident to v5, check for each such
			 * edge whether its other endpoint is adjacent to v1 to avoid double counting. */
			for(internal_node = START_OF_POOL(v5); internal_node < END_OF_POOL(v5); internal_node++){
				v4 = internal_node->id;

				// Consider only single edges.
				if(IS_NODE_OF_DOUBLE_EDGE(internal_node))
					continue;

				// This is a single edge. Add to count if the other endpoint is not in S and not adjacent to v1.
				if(NOT_IN_S(v4) && IS_NO_EDGE(v4, v1))
					count++;
			}
		}
	}

#if SAFETY_CHECKS
	// Compute count by brute force.
	u64 brute_count = M;

	// Only count simple edges.
	for(ii = START_OF_SINGLE_EDGES; ii < START_OF_DOUBLE_EDGES; ii += 2){
		// Ignore edges intersecting S.
		if(NOT_IN_S(points[ii]) && NOT_IN_S(points[ii + 1])){
			// Check both orientations of the edge.
			if(IS_NO_EDGE(v1,points[ii]) && IS_NO_EDGE(v3,points[ii + 1]))
				brute_count--;
			if(IS_NO_EDGE(v3,points[ii]) && IS_NO_EDGE(v1,points[ii + 1]))
				brute_count--;
		}
	}

	if(count != brute_count){
		LOG_ERR("ERROR: count %"PRIu64", brute count %"PRIu64".\n", count, brute_count);
		printAllNeighbors();
		finalize(safety_check_failure);
	}

	LOG_DEBUG("count of bad pairs was confirmed by brute count.\n");
#endif /* SAFETY_CHECKS */

	return M - count;
}

/* For a given simple 2-path S = v1-v2-v3, returns the number of simple 2-paths v4-v5-v6 which are vertex
 * disjoint from S, and such that (v1,v4), (v2,v5) and (v3,v6) are non-edges (labeling according to the paper).
 * Done by counting the "bad" 2-paths and subtracting them from P2, the total number of simple 2-paths.		 */
u64 computeBDG(int* S){
	int 	v1 = S[0], v2 = S[1], v3 = S[2], v4, v5, v6;
	int 	jj, single_neighbors, encountered_count = 0;
	u64 	ii, count = 0;
	adjNode *node, *internal_node;

	// First count paths v4-v5-v6 intersecting S.
	for(ii = 0; ii < 3; ii++){
		// Count paths such that v5 is in S (and maybe v4 or v6 as well).
		count += FALL_2(V[S[ii]].single_edges_num);

		// Count paths such that v5 not in S. Iterate all single neighbors v5 of S[ii].
		for(node = START_OF_POOL(S[ii]); node < END_OF_POOL(S[ii]); node++){
			v5 = node->id;

			// Consider only single edges to a vertex v5 not in S.
			if(IS_NODE_OF_SINGLE_EDGE(node) && NOT_IN_S(v5)){
				/* The path can continue on any single edge incident to v5 but the one backwards to S[ii].
				 * For each path both directions are counted (i.e. S[ii] can be either v4 or v6). 		*/
				count += 2 * (V[v5].single_edges_num - 1);

				// Avoid double counting if v4,v6 are both in S.
				for(jj = 0; jj < ii; jj++){
					if(IS_SINGLE_EDGE(v5,S[jj]))
						count -= 2;
				}
			}
		}
	}

	// Now count paths v4-v5-v6 disjoint from S, such that a forbidden edge exists.

	// First count paths such that (v2,v5) is an edge. Iterate all neighbors v5 of v2 not in S.
	for(node = START_OF_POOL(v2); node < END_OF_POOL(v2); node++){
		v5 = node->id;
		if(v5 != v1 && v5 != v3){
			// Count the single edges (v5,u) such that u is not in S.
			single_neighbors = V[v5].single_edges_num;
			for(jj = 0; jj < 3; jj++){
				if(IS_SINGLE_EDGE(v5, S[jj]))
					single_neighbors--;
			}
			// Any ordered pair of single neighbors forms a distinct bad 2-path.
			count += FALL_2(single_neighbors);
		}
	}

	// Now count paths v4-v5-v6 such that (v1,v4) is an edge but (v2,v5) is not. Iterate all neighbors v4 of v1 not in S.
	for(node = START_OF_POOL(v1); node < END_OF_POOL(v1); node++){
		v4 = node->id;
		if(v4 != v2 && v4 != v3){
			/* Now iterate all single neighbors v5 of v4 (not in S and not adjacent
			 * to v2), and for each count the single edges (v5,v6) for "new" v6. */
			for(internal_node = START_OF_POOL(v4); internal_node < END_OF_POOL(v4); internal_node++){
				v5 = internal_node->id;

				// Consider only single edges.
				if(IS_NODE_OF_DOUBLE_EDGE(internal_node))
					continue;

				// This is a single edge. Ignore if the other endpoint is in S or adjacent to v2.
				if(NOT_IN_S(v5) && IS_NO_EDGE(v2, v5)){
					// The path can continue on any single edge incident to v5 but the one to v4 and those to S.
					count += V[v5].single_edges_num - 1;
					for(ii = 0; ii < 3; ii++){
						if(IS_SINGLE_EDGE(v5,S[ii]))
							count--;
					}

					/* For later use, Keep count of the number of paths v1-v4-v5 such that
					 * (v4,v5) is a single edge, and v5 is not in S and not adjacent to v2. */
					C6_array[v5]++;

					// Keep track of the encountered vertices v5 for later initialization.
					if(C6_array[v5] == 1){
						encountered_count++;
						encountered[encountered_count] = v5;
					}
				}
			}
		}
	}


	// Finish by counting paths v4-v5-v6 such that (v3,v6) is an edge but not (v1,v4) or (v2,v5).

	// Iterate all neighbors v6 of v3 not in S.
	for(node = START_OF_POOL(v3); node < END_OF_POOL(v3); node++){
		v6 = node->id;
		if(v6 != v1 && v6 != v2){
			/* Now iterate all single neighbors v5 of v6 (not in S and not adjacent to v2),
			 * and for each count the single edges (v5,v4) for "new" v4 not adjacent to v1. */
			for(internal_node = START_OF_POOL(v6); internal_node < END_OF_POOL(v6); internal_node++){
				v5 = internal_node->id;

				// Consider only single edges.
				if(IS_NODE_OF_DOUBLE_EDGE(internal_node))
					continue;

				// This is a single edge. Ignore if the other endpoint is in S or adjacent to v2.
				if(NOT_IN_S(v5) && IS_NO_EDGE(v2, v5)){
					// The path can continue on any single edge incident to v5 but the one to v6 and those to S.
					count += V[v5].single_edges_num - 1;
					for(ii = 0; ii < 3; ii++){
						if(IS_SINGLE_EDGE(v5,S[ii]))
							count--;
					}

					/* It remains to subtract paths v4-v5-v6 such that both (v1,v4) and (v3,v6) are edges (which forms a C_6).
					 * Every such path was already counted once for each appropriate v4, as kept in C6_array. */
					count -= C6_array[v5];

					// Final adjustment: if (v1,v6) is an edge then the non-path v6-v5-v6 was just subtracted.
					if(IS_EDGE(v1,v6))
						count++;
				}
			}
		}
	}

	// Set all positive values in C6_array back to 0.
	while(encountered_count > 0){
		C6_array[encountered[encountered_count]] = 0;
		encountered_count--;
	}

#if SAFETY_CHECKS
	// Compute count by brute force.
	u64 brute_count = P2;

	// Only count simple edges (v4,v5).
	for(ii = 0; ii < START_OF_DOUBLE_EDGES; ii += 2){
		v4 = points[ii];
		v5 = points[ii + 1];

		// Ignore edges intersecting S.
		if(!NOT_IN_S(v4) || !NOT_IN_S(v5))
			continue;

		// Ignore edges (v4,v5) with one of the forbidden edges (v1,v4) or (v2,v5).
		if(IS_NO_EDGE(v1,v4) && IS_NO_EDGE(v2,v5)){
			// Iterate all neighbors v6 of v5.
			for(node = START_OF_POOL(v5); node < END_OF_POOL(v5); node++){
				v6 = node->id;
				// Subtract from the count if v6 is a single neighbor of v5 other than v4,
				// not in S, and not adjacent to v3.
				if(IS_NODE_OF_SINGLE_EDGE(node) && v6 != v4 && NOT_IN_S(v6) && IS_NO_EDGE(v3,v6))
					brute_count--;
			}
		}

		// Check the other orientation of the edge in the same manner.
		v4 = points[ii + 1];
		v5 = points[ii];
		if(IS_NO_EDGE(v1,v4) && IS_NO_EDGE(v2,v5)){
			for(node = START_OF_POOL(v5); node < END_OF_POOL(v5); node++){
				v6 = node->id;
				if(IS_NODE_OF_SINGLE_EDGE(node) && v6 != v4 && NOT_IN_S(v6) && IS_NO_EDGE(v3,v6))
					brute_count--;
			}
		}
	}


	if(count != brute_count){
		LOG_ERR("ERROR: count %"PRIu64", brute count %"PRIu64".\n", count, brute_count);
		printAllNeighbors();
		finalize(safety_check_failure);
	}

	LOG_DEBUG("count of bad 2-paths was confirmed by brute count.\n");
#endif /* SAFETY_CHECKS */

	return P2 - count;
}

// Attempts to remove all loops one after the other. Returns false upon f/b-rejection, and true otherwise.
bool noLoops(){
	int S[3];
	u64 blg;
	u64 blm0 = M2 - (8 * Delta * double_count) - (Delta * Delta * loop_count);
	u64 blm1 = M - (6 * Delta * Delta) + (4 * Delta);

	total_noloops++;
	total_loop_num += loop_count;

	while(loop_count > 0){
		// Attempt to perform an L-switch.
		total_L_switch_calls++;
		if(!performL(S)){
			LOG_DEBUG("Forward L-Switch failed.\n");
			loop_fwd_rej++;
			return false;
		}

		// Check backwards rejection.
		blm0 += Delta * Delta;
		blg = computeBLG(S);
		LOG_DEBUG("blm0 %"PRIu64", blm1 %"PRIu64", P2 %"PRIu64", blg %"PRIu64", ", blm0, blm1, P2, blg);
		if(checkBackwardsRejection(blm0, blm1, blg, P2)){
			LOG_DEBUG("Loop backwards rejection.\n");
			loop_bck_rej++;
			return false;
		}
		LOG_DEBUG("one loop iteration completed.\n");
	}

	LOG_DEBUG("All loops removed successfully!\n");
	return true;
}

// Attempts to remove all double edges one after the other. Returns false upon f/b-rejection, and true otherwise.
bool noDoubles(){
	int S[3];
	u64 bdg;
	u64 bdm0 = M2 - (8 * Delta * double_count);
	u64 bdm1 = M2 - (4 * double_count * (2 * Delta - 3)) - (3 * Delta * Delta * Delta);

	total_nodoubles++;
	total_double_edges_num += double_count;

	while(double_count > 0){
		// Attempt to perform a D-switch.
		total_D_switch_calls++;
		if(!performD(S)){
			LOG_DEBUG("Forward D-Switch failed.\n");
			double_fwd_rej++;
			return false;
		}

		// Check backwards rejection.
		bdm0 += 8 * Delta;
		bdm1 += 8 * Delta - 12;
		bdg = computeBDG(S);
		LOG_DEBUG("bdm0 %"PRIu64", bdm1 %"PRIu64", P2 %"PRIu64", bdg %"PRIu64", ", bdm0, bdm1, P2, bdg);
		if(checkBackwardsRejection(bdm0, bdm1, bdg, P2)){
			LOG_DEBUG("Double edge backwards rejection.\n");
			double_bck_rej++;
			return false;
		}
		LOG_DEBUG("one double edge iteration completed.\n");
	}

	LOG_DEBUG("All double edges removed successfully!\n");

	return true;
}

// Main program. Verifies input and configuration, generates random graphs, and outputs them.
int main(int argc, char** argv) {

	int 	ii, number_of_copies, input_length;
	bool	count_failed_attempts = CONSECUTIVE_FAILED_ATTEMPTS_LIMIT;
	nd_pair *input_pairs;

	int total_attempts = 0, generation_fails = 0, noLoops_fails = 0,
		noDoubles_fails = 0, consecutive_failed_attempts = 0;

	// First check if the program was run with the help option.
	if(argc == 2 && (!strcmp(argv[1], "h") || !strcmp(argv[1], "-h") || !strcmp(argv[1], "--h"))){
		printInputFormat();
		exit(user_termination);
	}

	// Check if the input is given via command line.
	bool is_input_in_file = checkInputType(argc, argv);

	// If files are used, verify the number of arguments given to the program is in range.
	if(is_input_in_file && (argc < 2 || argc > 4)){
		LOG_ERR("\nInvalid arguments given to the program!\n");
		printInputFormat();
		exit(input_failure);
	}

	// Open log file.
	char* log_file_name = (is_input_in_file && (argc == 4)) ? argv[3] : DEFAULT_LOG_FILE;
	log_file = fopen(log_file_name, "w");
	if(!log_file){
		LOG_ERR("\nCould not open log file. Exiting.\n");
		exit(file_failure);
	}

	// Open output file if provided.
	if(is_input_in_file && argc > 2){
		output_file = fopen(argv[2], "w");
		if(!output_file){
			LOG_ERR("\nCould not open output file. Exiting.\n");
			fclose(log_file);
			exit(file_failure);
		}
	}
	else{
		// stdout is the default output stream if no output file was provided.
		output_file = stdout;
	}

	// Verify there is not any faulty user configuration.
	checkConfiguration(argv);

	// Get input and verify it is valid.
	if(is_input_in_file){
		// Extract input from the input file.
		char **word_array = readInputFile(argv[1], &input_length);

		// Check input.
		input_pairs = verifyInput(word_array, input_length / 2, &number_of_copies);

		// Free the memory allocated by readInputFile.
		free(word_array[0]);
		free(word_array);
	}
	else{
		// Check input as stored in argv.
		input_length = argc - 1;
		input_pairs = verifyInput(argv + 1, input_length / 2, &number_of_copies);
	}

	// Allocate memory and create all data structures which only depend on the degree sequence itself.
	initialize(input_pairs, input_length / 2);

	if(DO_NOT_OUTPUT)
		LOG_INFO("\n---------- DO_NOT_OUTPUT flag is turned on. No graphs output. ----------\n\n");

	for(ii = 1; ii <= number_of_copies; ii++){
		LOG_DEBUG("\n\n------------------------------ Starting graph %d ------------------------------ \n\n", ii);
		if(SHOW_PROGRESS)
			LOG_ERR("%d\n", ii);

		// Loop until a simple graph is obtained.
		while(true){
			// Generate pairings until a valid one is obtained.
			while(true){
				// Alert user if many consecutive failed attempts occur.
				if(count_failed_attempts && consecutive_failed_attempts >= CONSECUTIVE_FAILED_ATTEMPTS_LIMIT){
					LOG_ERR("The last %d attempts to generate a graph failed.\n", consecutive_failed_attempts);
					PRINT_OVERALL_STATS(ERR);
					approveContinuation("Continue anyway (such message will not be prompted again)?");
					count_failed_attempts = false;
				}
				total_attempts++;
				if(generatePairing())
					break;
				generation_fails++;
				consecutive_failed_attempts++;
			}

			// Attempt to remove all loops if there are any. Start over upon failure.
			if(B1 > 0){
				if(loop_count > 0){
					if(!noLoops()){
						LOG_DEBUG("noLoops failed!\n");
						noLoops_fails++;
						consecutive_failed_attempts++;
						continue;
					}

					LOG_DEBUG("noLoops succeeded!\n");
				}
				else
					LOG_DEBUG("There are no loops to remove.\n");
			}

			verifyDataStructures();

			// Attempt to remove all double edges if there are any. Start over upon failure.
			if(B2 > 0){
				if(double_count > 0){
					if(!noDoubles()){
						LOG_DEBUG("noDoubles failed!\n");
						noDoubles_fails++;
						consecutive_failed_attempts++;
						continue;
					}

					LOG_DEBUG("noDoubles succeeded!\n");
				}
				else
					LOG_DEBUG("There are no double edges to remove.\n");
			}

			verifyDataStructures();
			LOG_DEBUG("Graph %d produced!\n", ii);
			consecutive_failed_attempts = 0;
			break;
		}

		// A simple graph has been obtained. Print it unless graphs output is disabled.
		outputGraph();
	}  /* End of main loop. */

	LOG_DEBUG("Finished producing all %d graphs.\n", number_of_copies);

	// Print statistics unless disabled.
	if(PRINT_SUMMARY){
		LOG_INFO("\n----------------------------------------------------------- Summary "
				 "-----------------------------------------------------------\n\n");

		LOG_INFO("Number of graphs = %d, n = %d, Delta = %" PRIu64 ", Loops limit = %d, Double edges limit = %d.\n",
				 number_of_copies, n, Delta, B1, B2);

		PRINT_OVERALL_STATS(INFO);
		printStatisticsPairings(generation_fails, total_attempts);
		printStatisticsLoops();
		printStatisticsDoubles();
	}

	finalize(successful_termination);
}
